/*---------------------------------------------------------------
 *	SAADCによるA/D変換の動作確認(μT-Kernel 3.0用)
 *
 *	Copyright (C) 2022-2024 by T3 WG of TRON Forum
 *---------------------------------------------------------------*/
#include <tk/tkernel.h>
#include <tm/tmonitor.h>

//-------- SAADCのレジスタ定義(※A) ------------------------
// SAADCのレジスタ定義
#define SAADC(r)		(SAADC_BASE + SAADC_##r)
#define SAADC_BASE		0x40007000

#define SAADC_TASKS_START	0x000
#define SAADC_TASKS_SAMPLE	0x004
#define SAADC_TASKS_STOP	0x008
#define SAADC_EVENTS_END	0x104
#define SAADC_EVENTS_DONE	0x108
#define SAADC_STATUS		0x400
#define SAADC_ENABLE		0x500

#define SAADC_PSELP(n)	 	(0x510 + (n) * 0x10)
#define SAADC_PSELN(n)	 	(0x514 + (n) * 0x10)
#define SAADC_CONFIG(n)		(0x518 + (n) * 0x10)

#define SAADC_RESOLUTION	0x5F0
#define SAADC_RESULT_PTR	0x62C
#define SAADC_RESULT_MAXCNT	0x630

//-------- LED制御のためのGPIOの初期設定 ------------------------
LOCAL void led_init(void)
{
	// COL1..COL5とROW1..ROW5に接続されたGPIOピンを出力に設定
	out_w(GPIO(P0, PIN_CNF(28)), 1);	// GPIO P0.28(COL1に接続)
	out_w(GPIO(P0, PIN_CNF(11)), 1);	// GPIO P0.11(COL2に接続)
	out_w(GPIO(P0, PIN_CNF(31)), 1);	// GPIO P0.31(COL3に接続)
	out_w(GPIO(P1, PIN_CNF(05)), 1);	// GPIO P1.05(COL4に接続)
	out_w(GPIO(P0, PIN_CNF(30)), 1);	// GPIO P0.30(COL5に接続)
	out_w(GPIO(P0, PIN_CNF(21)), 1);	// GPIO P0.21(ROW1に接続)
	out_w(GPIO(P0, PIN_CNF(22)), 1);	// GPIO P0.22(ROW2に接続)
	out_w(GPIO(P0, PIN_CNF(15)), 1);	// GPIO P0.15(ROW3に接続)
	out_w(GPIO(P0, PIN_CNF(24)), 1);	// GPIO P0.24(ROW4に接続)
	out_w(GPIO(P0, PIN_CNF(19)), 1);	// GPIO P0.19(ROW5に接続)
}

//-------- GPIOの特定のピンpinの出力をval(0または1)に設定 -------
//	pinで指定された以外のピンの状態は変化しない
LOCAL void out_gpio_pin(UW port, UW pin, UW val)
{
	INT port_addr;

	if(port == 1){							// P1の場合
		if(val)
			port_addr = GPIO(P1, OUTSET);	// P1でvalが1の場合
		else
			port_addr = GPIO(P1, OUTCLR);	// P1でvalが0の場合

	} else {								// P0の場合
		if(val)
			port_addr = GPIO(P0, OUTSET);	// P0でvalが1の場合
		else
			port_addr = GPIO(P0, OUTCLR);	// P0でvalが0の場合
	}
	out_w(port_addr, (1 << pin));			// 指定のピンに1を出力
}

//-------- LEDの制御(連載第7回のリスト4と同じ) (※L)-------------
// ROW1..ROW5のいずれか(rowで番号指定)に接続されたGPIOピンにvalを設定
LOCAL void out_row_gpio(UW row, UW val)
{
	switch(row){
	  case 1:
		out_gpio_pin(0, 21, val);		// GPIO P0.21(ROW1に接続)にvalを設定
		return;
	  case 2:
		out_gpio_pin(0, 22, val);		// GPIO P0.22(ROW2に接続)にvalを設定
		return;
	  case 3:
		out_gpio_pin(0, 15, val);		// GPIO P0.15(ROW3に接続)にvalを設定
		return;
	  case 4:
		out_gpio_pin(0, 24, val);		// GPIO P0.24(ROW4に接続)にvalを設定
		return;
	  case 5:
		out_gpio_pin(0, 19, val);		// GPIO P0.19(ROW5に接続)にvalを設定
		return;
	}
}

// rowで指定した行のみを点灯させるためのGPIOピンの設定
//	ROW1..ROW5に接続されたGPIOピンのうち、
//	rowで指定した1本のみを1に設定し、他は0に設定する
LOCAL void set_row_gpio(UW row)
{
	UW	cnt;
	for(cnt = 1; cnt <= 5; cnt++){
		out_row_gpio(cnt, ((cnt == row) ? 1 : 0));	// 指定の1本のみ1を設定
	}
}

//-------- LEDの制御(連載第11回での追加分) (※M)-----------------
// COL1..COL5のいずれか(colで番号指定)に接続されたGPIOピンにvalを設定
LOCAL void out_col_gpio(UW col, UW val)
{
	switch(col){
	  case 1:
		out_gpio_pin(0, 28, val);		// GPIO P0.28(COL1に接続)にvalを設定
		return;
	  case 2:
		out_gpio_pin(0, 11, val);		// GPIO P0.11(COL2に接続)にvalを設定
		return;
	  case 3:
		out_gpio_pin(0, 31, val);		// GPIO P0.31(COL3に接続)にvalを設定
		return;
	  case 4:
		out_gpio_pin(1,  5, val);		// GPIO P1.05(COL4に接続)にvalを設定
		return;
	  case 5:
		out_gpio_pin(0, 30, val);		// GPIO P0.30(COL5に接続)にvalを設定
		return;
	}
}

// colで指定した行のみを点灯させるためのGPIOピンの設定
//	COL1..COL5に接続されたGPIOピンのうち、
//	colで指定した1本のみを0に設定し、他は1に設定する
LOCAL void set_col_gpio(UW col)
{
	UW	cnt;
	for(cnt = 1; cnt <= 5; cnt++){
		out_col_gpio(cnt, ((cnt == col) ? 0 : 1));	// 指定の1本のみ0を設定
	}
}

//-------- SAADCを使ったA/D変換 ---------------------------------
// AD変換の結果を入れるメモリ
#define	SAADC_DATACNT	2
static _H saadc_data[SAADC_DATACNT];

INT	adc_poll_cnt;			// A/D変換の終了までのポーリング回数

// 2チャンネルのアナログ入力値をデジタル値に変換(※B)
LOCAL void analogRead2(INT ch0_ain, INT ch1_ain){

	out_w(SAADC(PSELP(0)), ch0_ain);	// CH[0]のP入力の設定(※C)
	out_w(SAADC(CONFIG(0)), 0);			// CH[0]のCONFIGの設定
	
	out_w(SAADC(PSELP(1)), ch1_ain);	// CH[1]のP入力の設定(※C)
	out_w(SAADC(CONFIG(1)), 0);			// CH[1]のCONFIGの設定

	out_w(SAADC(RESULT_PTR), (UW)saadc_data);	// 結果を入れるRAMアドレス(※D)
	out_w(SAADC(RESULT_MAXCNT), SAADC_DATACNT);	// 結果を入れるRAMのデータ数

	out_w(SAADC(ENABLE), 1);			// SAADCのイネーブル(※E)
	out_w(SAADC(RESOLUTION), 0);		// resolution=8bitの設定(※F)

	out_w(SAADC(EVENTS_END), 0);		// 書込完了のフラグをクリア(※G)
	out_w(SAADC(TASKS_START), 1);		// A/D変換の開始(※H)
	out_w(SAADC(TASKS_SAMPLE), 1);		// サンプル入力
	adc_poll_cnt = 0;

	for(;;){							// A/D変換の書込完了を待つループ
		if(in_w(SAADC(EVENTS_END)))		// 書込完了のレジスタを確認(※J)
			 return;
		adc_poll_cnt++;					// ポーリング回数をプラス1
		tk_dly_tsk(10);
	}
}

//----------------------------------------------------------
LOCAL W cur_time(void)			// ミリ秒単位の現在時刻の下位32ビットを返す
{
	SYSTIM	tim;
	tk_get_tim(&tim);			// 現在時刻を取得
	return(tim.lo);
}

LOCAL INT scale_1to5(INT in_data){	// 0..255の範囲のin_dataを1..5に変換(※K)
	INT out_data = in_data / (256 / 5) + 1;
	if(out_data < 1) out_data = 1;
	if(out_data > 5) out_data = 5;
	return(out_data);
}

//----------------------------------------------------------
EXPORT void usermain( void )				// (※N)
{
	W itime = cur_time();					// プログラム開始時刻
	led_init();
	
	// 無限ループでA/D変換を連続実行
	for(;;){
		analogRead2(1, 2);	// ch0=AIN0(RING0),ch1=AIN1(RING1) でA/D変換

		// ch0:入力値小→左方のCOL1を点灯、大→右方のCOL5を点灯(※P)
		INT ch0_1to5 = scale_1to5(saadc_data[0]);

		// ch1:入力値小→下方のROW5、入力値大→上方のROW1を点灯、上下反転(※R)
		INT ch1_1to5 = 6 - scale_1to5(saadc_data[1]);

		// A/D変換結果をコンソールに出力(※S)
		tm_printf("%6d: poll=%d, ch0=%4d (COL%d), ch1=%4d (ROW%d)\n",
			(cur_time() - itime), adc_poll_cnt,
			saadc_data[0], ch0_1to5, saadc_data[1], ch1_1to5);

		set_col_gpio(ch0_1to5);	// ch0の入力値を示すCOL位置のLEDを点灯(※Q)
		set_row_gpio(ch1_1to5);	// ch1の入力値を示すROW位置のLEDを点灯

		tk_dly_tsk(100);
	}
    tk_slp_tsk(TMO_FEVR);
}
